// =============================================================================
// Stereo recognition - determining depth in scene
// Image Processing and Computer Vision (COMS 30121), assignment 2
//
// author: Jaroslav Borovicka, December 2002 - January 2003
//         University of Bristol
//         Czech Technical University in Prague
// =============================================================================
// matrices.cpp
// matrix and vector classes
// =============================================================================

// This small library defines the classes for vector and matrix calculus
// (i.e. data structures and operators) required for the assignment.
//
// vector, matrix - float classes
// ivector, imatrix - integer classes 

#include "matrices.h"

// =============================================================================
// INT32 **int32_matrix(int rows, int columns)
//
// Creates a 2D matrix of size rows*columns
//
// =============================================================================

INT32 **int32_matrix(int rows, int columns)
{
    INT32 **mat;

    mat = (INT32 **) calloc(rows, sizeof(INT32 *));
    if (mat == NULL)
    {
      fprintf(stderr,"Couldn't create row vector.\n");
      return NULL;
    }
    for (int i = 0; i < rows; i++)
    {
      mat[i] = (INT32 *) calloc(columns, sizeof(INT32));
      if (mat[i] == NULL)
      {
        fprintf(stderr,"Couldn't create one of the rows.\n");
        for (int j = 0; j < i; j++) free (mat[j]);
        free (mat);
        return NULL;
      }
    }

    return mat;
}

// =============================================================================
//  void free_int32_matrix(int32 **mat, int rows)
//
// =============================================================================

void free_int32_matrix(INT32 **mat, int rows)
{
  if (mat == NULL) {
    return;
  }

  for (int i = 0; i < rows; i++) free (mat[i]);
  free (mat);

  return;

}


// =============================================================================
// float **float_matrix(int rows, int columns)
//
// Creates a 2D matrix of size rows*columns
//

float **float_matrix(int rows, int columns)
{
    float **mat;

    mat = (float **) calloc(rows, sizeof(float *));
    if (mat == NULL)
    {
      fprintf(stderr,"Couldn't create row vector.\n");
      return NULL;
    }
    for (int i = 0; i < rows; i++)
    {
      mat[i] = (float *) calloc(columns, sizeof(float));
      if (mat[i] == NULL)
      {
        fprintf(stderr,"Couldn't create one of the rows.\n");
        for (int j = 0; j < i; j++) free (mat[j]);
        free (mat);
        return NULL;
      }
    }

    return mat;
}

// =============================================================================
//  void free_float_matrix(float **p, int height)

void free_float_matrix(float **mat, int rows)
{
  if (mat == NULL) {
    return;
  }

  for (int i = 0; i < rows; i++) free (mat[i]);
  free (mat);

  return;

}

// =============================================================================
//
// Vector method declarations
//
// =============================================================================

vector::vector() { m = 0; data = NULL; }

vector::vector(int a)
{
  m = a;
  data = (float *) calloc(m, sizeof(float));
}

vector::~vector() { if (this != NULL) free(data); }

vector operator *(float c, vector v)
{
  // allocate memory for the result
  vector result(v.m);

  for (int k = 0; k < v.m; k++)
    result.data[k] = c * v.data[k];

  return result;
}

float operator *(vector v1, vector v2)
{
  // check correctness of sizes
  if (v1.m != v2.m) return NULL;

  // allocate memory for the result
  float result = 0;

  for (int k = 0; k < v1.m; k++)
    result += v1.data[k] * v2.data[k];

  return result;
}

vector operator ^(vector u, vector v)  // vector multiplication
{
  // check correctness of sizes
  if ((u.m != 3) || (v.m != 3)) return NULL;

  vector r(3);
  // allocate memory for the result
  r.data[0] = u.data[1] * v.data[2] - u.data[2] * v.data[1]; 
  r.data[1] = u.data[2] * v.data[0] - u.data[0] * v.data[2];
  r.data[2] = u.data[0] * v.data[1] - u.data[1] * v.data[0];

  return r;
}

vector operator +(vector v1, vector v2)
{
  // check correctness of sizes
  if (v1.m != v2.m) return vector();

  // allocate memory for the result
  vector result(v1.m);

  for (int k = 0; k < v1.m; k++)
    result.data[k] = v1.data[k] + v2.data[k];

  return result;
}

vector operator -(vector v1, vector v2)
{
  // check correctness of sizes
  if (v1.m != v2.m) return vector();

  // allocate memory for the result
  vector result(v1.m);

  for (int k = 0; k < v1.m; k++)
    result.data[k] = v1.data[k] - v2.data[k];

  return result;
}

vector vector::operator-()
{
  vector result(this->length());
  for (int i = 1; i <= result.length(); i++) result[i] = -(*this)[i];
  return result;
}

vector::vector(const vector &v)   // copy constructor
{
  m = v.m;
  data = (float *) calloc (m, sizeof(float));
  memcpy (data, v.data, m * sizeof(float));
}

vector& vector::operator =(const vector &v)
{
  if (this == &v) return *this;
  free(data);
  m = v.m;
  data = (float *) calloc (m, sizeof(float));
  memcpy (data, v.data, m * sizeof(float));

  return *this;
}

int operator==(vector v1, vector v2)
{
  if (v1.m != v2.m) return 0;
  for (int i = 0; i < v1.m; i++)
    if (v1.data[i] != v2.data[i]) return 0;

  return 1;
}

int operator!=(vector v1, vector v2)
{
  if (v1.m != v2.m) return 1;
  for (int i = 0; i < v1.m; i++)
    if (v1.data[i] != v2.data[i]) return 1;

  return 0;
}

void vector::setdata(float *content, int l)
{
//  free (data);
  data = content;
  m = l;
};

float vector::abs()
{
  float res = 0;
  for (int i = 0; i < m; i++)
    res += data[i]*data[i];
  res = pow (res, 0.5);
  return res;
}

void vector::print()
{
  for (int i = 0; i < m; i++)
  {
    printf ("%f ", data[i]);
  }
  printf ("\n");
}

// =============================================================================
//
// Matrixrow method declarations
//
// =============================================================================

matrixrow::matrixrow (const matrixrow &mr)   // copy constructor
{
  m = mr.m;
  data = mr.data;
}

matrixrow matrixrow::operator =(vector v)
{
  for (int i = 1; i <= m; i++) data[i-1] = v[i];
//  memcpy (this->data, v.data, v.length() * sizeof(INT32));
  return *this;
}

matrixrow matrixrow::operator =(matrixrow v)
{
  for (int i = 1; i <= m; i++) data[i-1] = v[i];
//  memcpy (this->data, v.data, v.length() * sizeof(INT32));
  return *this;
}

// =============================================================================
//
// Matrix method declarations
//
// =============================================================================

matrix::matrix() { m = 0; n = 0; data = NULL; };

matrix::matrix(int a, int b)
{
  m = a; n = b;
  data = float_matrix(m, n);
  if (data == NULL) { m = 0; n = 0; }
}

matrix::~matrix() { if (this != NULL) free_float_matrix(data, m); };

matrix operator *(float c, matrix mat)
{
  // allocate memory for the result
  matrix result(mat.m, mat.n);

  for (int i = 0; i < mat.m; i++)
    for (int j = 0; j < mat.n; j++)
      result.data[i][j] = c * mat.data[i][j];

  return result;
}

matrix operator *(matrix m1, matrix m2)
{
  // check correctness of sizes
  if (m1.n != m2.m) return matrix();

  // allocate memory for the result
  matrix result(m1.m, m2.n);

  for (int i = 0; i < m1.m; i++)
  {
    for (int j = 0; j < m2.n; j++)
    {
      result.data[i][j] = 0;
      for (int k = 0; k < m1.n; k++)
      {
        result.data[i][j] += m1.data[i][k] * m2.data[k][j];
      }
    }
  }
  return result;
};

vector operator *(matrix mat, vector vect)
{
  // check correctness of sizes
  if (mat.n != vect.length()) return vector();

  // allocate memory for the result
  vector result(mat.m);

  for (int i = 1; i <= mat.m; i++)
  {
    result[i] = 0;
    for (int k = 1; k <= mat.n; k++)
    {
      result[i] += mat[i][k] * vect[k];
    }
  }
  return result;
};

vector operator *(vector vect, matrix mat)
{
  // check correctness of sizes
  if (mat.m != vect.length()) return vector();

  // allocate memory for the result
  vector result(mat.n);

  for (int i = 1; i <= mat.n; i++)
  {
    result[i] = 0;
    for (int k = 1; k <= mat.m; k++)
    {
      result[i] += vect[k] * mat[k][i];
    }
  }
  return result;
};

matrix operator +(matrix m1, matrix m2)
{
  if ((m1.m != m2.m) || (m1.n != m2.n)) return matrix();

  matrix result(m1.m, m1.n);
  for (int i = 0; i < m1.m; i++)
  {
    for (int j = 0; j < m1.n; j++)
      result.data[i][j] = m1.data[i][j] + m2.data[i][j];
  }
  return result;
}

matrix operator -(matrix m1, matrix m2)
{
  if ((m1.m != m2.m) || (m1.n != m2.n)) return matrix();

  matrix result(m1.m, m1.n);
  for (int i = 0; i < m1.m; i++)
  {
    for (int j = 0; j < m1.n; j++)
      result.data[i][j] = m1.data[i][j] - m2.data[i][j];
  }
  return result;
}


matrix::matrix(const matrix &mat)   // copy constructor
{
  m = mat.m;
  n = mat.n;
  data = float_matrix(m, n);
  for (int i = 0; i < m; i++)
    memcpy (data[i], mat.data[i], n * sizeof(float));
}

matrix& matrix::operator =(const matrix &mat)
{
  if (this == &mat) return *this;
  free_float_matrix(data, m);
  m = mat.m;
  n = mat.n;

  if ((m == 0) || (n == 0))
  {
    data = NULL;
    return *this;
  }
  
  data = float_matrix(m, n);
  for (int i = 0; i < m; i++)
    memcpy (data[i], mat.data[i], n * sizeof(float));

  return *this;
}


matrix matrix::row(int i)
{
  matrix r(1,n);
  for (int j = 0; j < n; j++) r[1][j+1] = data[i-1][j];
//  memcpy (r.data, data[i-1], n * sizeof(float));
  return r;
}

matrix matrix::column(int j)
{
  matrix c(m,1);
  for (int i = 0; i < m; i++) c.data[i][0] = data[i][j-1];
  return c;
}

vector matrix::vrow(int i)
{
  vector r(n);
  for (int j = 0; j < n; j++) r[j+1] = data[i-1][j];
  return r;
}

const vector matrix::vrow(int i) const
{
  vector r(n);
  for (int j = 0; j < n; j++) r[j+1] = data[i-1][j];
  return r;
}


vector matrix::vcolumn(int j)
{
  vector v(m);
  for (int i = 0; i < m; i++) v[i+1] = data[i][j-1];
  return v;
}

void matrix::setdata(float **content, int a, int b)
{
  free_float_matrix (data, m);
  data = content;
  m = a;
  n = b;
};

matrix matrix::transpose()
{
  matrix mt(n,m);
  for (int i = 0; i < m; i++)
  {
    for (int j = 0; j < n; j++)
    {
      mt.data[j][i] = data[i][j];
    }
  }
  return mt;
}

matrix matrix::inverse ()
{
  if (m != n) return matrix();

  int i, j, k;
  matrix mt (m, m);
  matrix hlp = *this;
  float a, b;

  for (i = 0; i < m; i++) mt.data[i][i] = 1;

  for (j = 0; j < m - 1; j++)
  {
    a = hlp.data[j][j];
    for (i = j + 1; i < m; i++)
    {
      b = hlp.data[i][j];
      for (k = j; k < m; k++)
      {
        hlp.data[i][k] = hlp.data[i][k] * a - hlp.data[j][k] * b;
      }
      for (k = 0; k < m; k++)
      {
        mt.data[i][k] = mt.data[i][k] * a - mt.data[j][k] * b;
      }
    }
  }

  for (j = m - 1; j >= 0; j--)
  {
    // if error, return empty matrix;
    if (hlp.data[j][j] == 0)
    {
      printf ("cannot invert.\n");
      return matrix(m, m);
    }
    for (k = 0; k < m; k++)
    {
      mt.data[j][k] = mt.data[j][k] / hlp.data[j][j];
    }
    hlp.data[j][j] = 1;

    for (i = j - 1; i >= 0; i--)
    {
      b = hlp.data[i][j];
      hlp.data[i][j] = 0;

      for (k = 0; k < m; k++)
      {
        mt.data[i][k] = mt.data[i][k] - mt.data[j][k] * b;
      }
    }
  }
  return mt;
}

void matrix::print()
{
  for (int i = 0; i < m; i++)
  {
    printf ("Row %d: ", i+1);
    for (int j = 0; j < n; j++)
    {
      printf ("%f ", data[i][j]);
    }
    printf ("\n");
  }
  return;
}

// =============================================================================
//
// Integer Vector method declarations
//
// =============================================================================

ivector::ivector() { m = 0; data = NULL; }

ivector::ivector(int a)
{
  m = a;
  data = (INT32 *) calloc(m, sizeof(INT32));
}

ivector::~ivector() { if (this != NULL) free(data); }

ivector operator *(INT32 c, ivector v)
{
  // allocate memory for the result
  ivector result(v.m);

  for (int k = 0; k < v.m; k++)
    result.data[k] = c * v.data[k];

  return result;
}

INT32 operator *(ivector v1, ivector v2)
{
  // check correctness of sizes
  if (v1.m != v2.m) return NULL;

  // allocate memory for the result
  INT32 result = 0;

  for (int k = 0; k < v1.m; k++)
    result += v1.data[k] * v2.data[k];

  return result;
}

ivector operator ^(ivector u, ivector v)  // vector multiplication
{
  // check correctness of sizes
  if ((u.m != 3) || (v.m != 3)) return NULL;

  ivector r(3);
  // allocate memory for the result
  r.data[0] = u.data[1] * v.data[2] - u.data[2] * v.data[1]; 
  r.data[1] = u.data[2] * v.data[0] - u.data[0] * v.data[2];
  r.data[2] = u.data[0] * v.data[1] - u.data[1] * v.data[0];

  return r;
}

ivector operator +(ivector v1, ivector v2)
{
  // check correctness of sizes
  if (v1.m != v2.m) return ivector();

  // allocate memory for the result
  ivector result(v1.m);

  for (int k = 0; k < v1.m; k++)
    result.data[k] = v1.data[k] + v2.data[k];

  return result;
}

ivector operator -(ivector v1, ivector v2)
{
  // check correctness of sizes
  if (v1.m != v2.m) return ivector();

  // allocate memory for the result
  ivector result(v1.m);

  for (int k = 0; k < v1.m; k++)
    result.data[k] = v1.data[k] - v2.data[k];

  return result;
}

int operator &(ivector v1, ivector v2)
{
  if (v1.length() != v2.length()) return -1;
  int d = 0;
  for (int i = 1; i <= v1.length(); i++)
    if (v1[i] != v2[i]) d++;

  return d;
}

ivector ivector::operator-()
{
  ivector result(this->length());
  for (int i = 1; i <= result.length(); i++) result[i] = -(*this)[i];
  return result;
}

ivector::ivector(const ivector &v)   // copy constructor
{
  m = v.m;
  data = (INT32 *) calloc (m, sizeof(INT32));
  memcpy (data, v.data, m * sizeof(INT32));
}

ivector& ivector::operator =(const ivector &v)
{
  if (this == &v) return *this;
  free(data);
  m = v.m;
  data = (INT32 *) calloc (m, sizeof(INT32));
  memcpy (data, v.data, m * sizeof(INT32));

  return *this;
}

int operator==(ivector v1, ivector v2)
{
  if (v1.m != v2.m) return 0;
  for (int i = 0; i < v1.m; i++)
    if (v1.data[i] != v2.data[i]) return 0;

  return 1;
}

int operator!=(ivector v1, ivector v2)
{
  if (v1.m != v2.m) return 1;
  for (int i = 0; i < v1.m; i++)
    if (v1.data[i] != v2.data[i]) return 1;

  return 0;
}

void ivector::setdata(INT32 *content, int l)
{
//  free (data);
  data = content;
  m = l;
};

INT32 ivector::abssq()
{
  INT32 res = 0;
  for (int i = 0; i < m; i++)
    res += data[i]*data[i];
  return res;
}

void ivector::print()
{
  for (int i = 0; i < m; i++)
  {
    printf ("%d ", data[i]);
  }
  printf ("\n");
}

// =============================================================================
//
// iMatrixrow method declarations
//
// =============================================================================

imatrixrow::imatrixrow (const imatrixrow &mr)   // copy constructor
{
  m = mr.m;
  data = mr.data;
}

imatrixrow imatrixrow::operator =(ivector v)
{
  for (int i = 1; i <= m; i++) data[i-1] = v[i];
//  memcpy (this->data, v.data, v.length() * sizeof(INT32));
  return *this;
}

imatrixrow imatrixrow::operator =(imatrixrow v)
{
  for (int i = 1; i <= m; i++) data[i-1] = v[i];
//  memcpy (this->data, v.data, v.length() * sizeof(INT32));
  return *this;
}


// =============================================================================
//
// Integer Matrix method declarations
//
// =============================================================================

imatrix::imatrix() { m = 0; n = 0; data = NULL; };

imatrix::imatrix(int a, int b)
{
  m = a; n = b;
  data = int32_matrix(m, n);
  if (data == NULL) { m = 0; n = 0; }
}

imatrix::~imatrix() { if (this != NULL) free_int32_matrix(data, m); };

imatrix operator *(imatrix m1, imatrix m2)
{
  // check correctness of sizes
  if (m1.n != m2.m) return imatrix();

  // allocate memory for the result
  imatrix result(m1.m, m2.n);

  for (int i = 0; i < m1.m; i++)
  {
    for (int j = 0; j < m2.n; j++)
    {
      result.data[i][j] = 0;
      for (int k = 0; k < m1.n; k++)
      {
        result.data[i][j] += m1.data[i][k] * m2.data[k][j];
      }
    }
  }
  return result;
};

ivector operator *(imatrix mat, ivector vect)
{
  // check correctness of sizes
  if (mat.n != vect.length()) return ivector();

  // allocate memory for the result
  ivector result(mat.m);

  for (int i = 1; i <= mat.m; i++)
  {
    result[i] = 0;
    for (int k = 1; k <= mat.n; k++)
    {
      result[i] += mat[i][k] * vect[k];
    }
  }
  return result;
};

ivector operator *(ivector vect, imatrix mat)
{
  // check correctness of sizes
  if (mat.m != vect.length()) return ivector();

  // allocate memory for the result
  ivector result(mat.n);

  for (int i = 1; i <= mat.n; i++)
  {
    result[i] = 0;
    for (int k = 1; k <= mat.m; k++)
    {
      result[i] += vect[k] * mat[k][i];
    }
  }
  return result;
};


imatrix operator +(imatrix m1, imatrix m2)
{
  if ((m1.m != m2.m) || (m1.n != m2.n)) return imatrix();

  imatrix result(m1.m, m1.n);
  for (int i = 0; i < m1.m; i++)
  {
    for (int j = 0; j < m1.n; j++)
      result.data[i][j] = m1.data[i][j] + m2.data[i][j];
  }
  return result;
}

imatrix operator -(imatrix m1, imatrix m2)
{
  if ((m1.m != m2.m) || (m1.n != m2.n)) return imatrix();

  imatrix result(m1.m, m1.n);
  for (int i = 0; i < m1.m; i++)
  {
    for (int j = 0; j < m1.n; j++)
      result.data[i][j] = m1.data[i][j] - m2.data[i][j];
  }
  return result;
}


imatrix::imatrix(const imatrix &mat)   // copy constructor
{
  m = mat.m;
  n = mat.n;
  data = int32_matrix(m, n);
  for (int i = 0; i < m; i++)
    memcpy (data[i], mat.data[i], n * sizeof(INT32));
}

imatrix& imatrix::operator =(const imatrix &mat)
{
  if (this == &mat) return *this;
  free_int32_matrix(data, m);
  m = mat.m;
  n = mat.n;
  if ((m == 0) || (n == 0))
  {
    data = NULL;
  }
  else
  {
    data = int32_matrix(m, n);
    for (int i = 0; i < m; i++)
      memcpy (data[i], mat.data[i], n * sizeof(INT32));
  }

  return *this;
}


imatrix imatrix::row(int i)
{
  imatrix r(1,n);
  for (int j = 0; j < n; j++) r[1][j+1] = data[i-1][j];
//  memcpy (r.data[1], data [i-1], n * sizeof(INT32));
  return r;
}

imatrix imatrix::column(int j)
{
  imatrix c(m,1);
  for (int i = 0; i < m; i++) c.data[i][0] = data[i][j-1];
  return c;
}

ivector imatrix::vrow(int i)
{
  ivector r(n);
  for (int j = 0; j < n; j++) r[j+1] = data[i-1][j];
  return r;
}

const ivector imatrix::vrow(int i) const
{
  ivector r(n);
  for (int j = 0; j < n; j++) r[j+1] = data[i-1][j];
  return r;
}


ivector imatrix::vcolumn(int j)
{
  ivector v(m);
  for (int i = 0; i < m; i++) v[i+1] = data[i][j-1];
  return v;
}

void imatrix::setdata(INT32 **content, int a, int b)
{
  free_int32_matrix (data, m);
  data = content;
  m = a;
  n = b;
};

imatrix imatrix::transpose()
{
  imatrix mt(n,m);
  for (int i = 0; i < m; i++)
  {
    for (int j = 0; j < n; j++)
    {
      mt.data[j][i] = data[i][j];
    }
  }
  return mt;
}

void imatrix::print()
{
  for (int i = 0; i < m; i++)
  {
    printf ("Row %d: ", i+1);
    for (int j = 0; j < n; j++)
    {
      printf ("%d ", data[i][j]);
    }
    printf ("\n");
  }
  return;
}
