// =============================================================================
// Stereo recognition - determining depth in scene
// Image Processing and Computer Vision (COMS 30121), assignment 2
//
// author: Jaroslav Borovicka, December 2002 - January 2003
//         University of Bristol
//         Czech Technical University in Prague
// =============================================================================
// stereo.cpp
// main program and interface
// =============================================================================

// note: matrix calculations are based on the classes 'vector' and 'matrix'
// which I define in the matrices.cpp and matrices.h files

#include "stereo.h"

#define MAX(a,b) ((a)>(b) ? (a) : (b))
#define MIN(a,b) ((a)<(b) ? (a) : (b))
#define ABS(a) ((a)>0 ? (a) : -(a))
#define SIGN(a) ((a)>0 ? 1 : -1)


// =============================================================================
// define a "Mexican hat" filter for further use as second derivative filter

  float mexhatsmall[5][5] =
    {{ 0, 0,-1, 0, 0},
     { 0,-1,-2,-1, 0},
     {-1,-2,16,-2,-1},
     { 0,-1,-2,-1, 0},
     { 0, 0,-1, 0, 0}};

// =============================================================================
// TPGMImage class methods
// =============================================================================

// =============================================================================
// TPGMImage::TPGMImage (char *f, char c)
//
// Constructor for the image class - loads the image, and copies the data
// to a float array.
//
// Parameters: char *f - filename of the image
//             char c  - either L or R according to which view to image
//                       represents
// =============================================================================

TPGMImage::TPGMImage (char *f, char c)
{
  index = c;
  Filename = strdup (f);
  Data = readpgm (Filename, &w, &h);    // load the image

  // copy the image data to the float array
  fData = float_array (w, h);
  for (int i = 0; i < h; i++)
    for (int j = 0; j < w; j++) fData[i][j] = Data[i][j];

  fEdgeMap = NULL;
  EdgeMap = NULL;
  fDirectionMap = NULL;
  DirectionMap = NULL;

}

// =============================================================================
// TPGMImage::~TPGMImage () - destructor
// =============================================================================

TPGMImage::~TPGMImage ()
{
  free (Filename);
  if (fData != NULL) free_float_array (fData, h);
  if (Data != NULL) free_uchar_array (Data, h);
  if (fEdgeMap != NULL) free_float_array (fEdgeMap, h);
  if (EdgeMap != NULL) free_uchar_array (EdgeMap, h);
  if (fDirectionMap != NULL) free_float_array (fDirectionMap, h);
  if (DirectionMap != NULL) free_uchar_array (DirectionMap, h);
}

// =============================================================================
// float TPGMImage::GetEdgeMap (PParams Params)
//
// First, if set from the command line, smoothes the image. Then the edge map
// is calculated using the convolution of the image with the horizontal and
// vertical Sobel filter, and merging the two resulting edge maps together,
// using L2 (Eucledian) norm.
//
// Parameters: PParams Params - command-line parameters
//
// Returns:    maximum value in the merged edge map (useful for normalising)
// =============================================================================

void TPGMImage::GetEdgeMap (PParams Params)
{

  float **sfh, **sfv;        // Sobel filters
  float **imageh, **imagev;  // horizontal and vertical edge maps


  float maxval;
  int i,j;

  char fn[40];

  // ---------------------------------------------------------------------------
  // first, smooth the image to get rid of the noise (if chosen);

  if (Params->smoothimage)
  {
     fData = smoothimage (fData, w, h, Params);
  }

  // ---------------------------------------------------------------------------
  // define Sobel filters

  sfh = float_array(3,3);
  sfv = float_array(3,3);

  sfv[0][0] = sfv[2][0] = -1;
  sfv[1][0] = -2;
  sfv[0][2] = sfv[2][2] = 1;
  sfv[1][2] = 2;

  sfh[0][0] = sfh[0][2] = 1;
  sfh[0][1] = 2;
  sfh[2][0] = sfh[2][2] = -1;
  sfh[2][1] = -2;

  // ---------------------------------------------------------------------------
  // apply Sobel filters

  printf ("Applying horizontal Sobel filter.\n");
  imageh = convolution (fData, w, h, sfh, 3, 3);
  printf ("Applying vertical Sobel filter.\n");
  imagev = convolution (fData, w, h, sfv, 3, 3);

  // save the filter files
  if (Params->saveall)
  {
    sprintf (fn, "output/sobelhoriz%c.pgm", index);
    writefloatpgm (imageh, w, h, fn, 0);
    sprintf (fn, "output/sobelvert%c.pgm", index);
    writefloatpgm (imagev, w, h, fn, 0);
  }

  // ---------------------------------------------------------------------------
  // merging the two sobel filters
  // (ignore border areas which are distorted because of
  //  using the smoothing filter)

  printf ("Merging the Sobel filters.\n");
  maxval = 0.0;
  EdgeMap = uchar_array (w, h);
  fEdgeMap = float_array (w, h);
  DirectionMap = uchar_array (w, h);
  fDirectionMap = float_array (w, h);

  for (i = Params->smoothimage + 1; i < h - Params->smoothimage - 1; i++)
  {
    for (j = Params->smoothimage + 1; j < w - Params->smoothimage - 1; j++)
    {
      fEdgeMap[i][j] =  pow( imageh[i][j] * imageh[i][j] +
                             imagev[i][j] * imagev[i][j], 0.5);
      if (fEdgeMap[i][j] > maxval) maxval = fEdgeMap[i][j];
    }
  }

  if (Params->saveall)
  {
    sprintf (fn, "output/sobelmerged%c.pgm", index);
    writefloatpgm (fEdgeMap, w, h, fn, 0);
  }

  // ---------------------------------------------------------------------------
  // run the 2nd derivative edge detector

  if (Params->use2ndderivative)
  {
    EdgesUsing2ndDer (Params);
  }

  // ---------------------------------------------------------------------------
  // thresholding and finding the direction map

  for (i = 0; i < h; i++)
  {
    for (j = 0; j < w; j++)
    {
      if (fEdgeMap[i][j]/maxval*255 < Params->contthreshold)
      {
        // the pixel does not correspond to an edge - set brightness to 0
        fEdgeMap[i][j] = 0;
        EdgeMap[i][j] = 0;
        fDirectionMap[i][j] = 0;
        DirectionMap[i][j] = 0;
      }
      else
      {
        // the pixel corresponds to an edge - calculate the direction
        fEdgeMap[i][j] = fEdgeMap[i][j] / maxval * 255.0;
        EdgeMap[i][j] = floor (fEdgeMap[i][j]);
        fDirectionMap[i][j] = atan2(imageh[i][j], imagev[i][j]);
        if (fDirectionMap[i][j] > M_PI_2) fDirectionMap[i][j] -= M_PI;
        if (fDirectionMap[i][j] < -M_PI_2) fDirectionMap[i][j] += M_PI;
        DirectionMap[i][j] =(unsigned char)((DirectionMap[i][j] / M_PI + 0.5) * 255) ;
      }
    }
  }

  if (Params->saveall)
  {
    // save intermediate files, if requested
    sprintf (fn, "output/edgemap%c.pgm", index);
    writepgm (EdgeMap, w, h, fn);
//    writepgm (dirmapuchar, w, h, "_edgedirections.pgm");
  }

  free_float_array (sfh, 3);
  free_float_array (sfv, 3);
  free_float_array (imageh, h);
  free_float_array (imagev, h);

  return;
}


// =============================================================================
// void TPGMImage::EdgesUsing2ndDer ()
//
// Uses the "second derivative" filter to make the edge detector more robust.
// The function convolutes the image with the "Mexican hat"
// filter. The Mexican hat filter is the second derivative of the Gaussian
// filter, and convoluting the image with this filter means looking up
// the second derivative of a smoothed image. For more mathematical
// background see report. The second derivative map is then examined
// with a simple zero-crossing filter. All edges in the edgemap
// which do not have support in a zero crossing of the second derivative,
// are then removed.
//
//  Returns:   nothing
// =============================================================================

void TPGMImage::EdgesUsing2ndDer (PParams P)
{
  float **secondder;  // second derivative map
  float **f;          // "mexican hat" filter
  int i, j, removeedge;
  char fn[40];

  printf ("Using the 2nd derivative filter.\n");

  // define the filter
  f = float_array(5,5);
  for (i = 0; i < 5; i++) for (j = 0; j < 5; j++) f[i][j] = mexhatsmall[i][j];

  secondder = convolution (fEdgeMap, w, h, f, 5, 5);

  // now look for the zero crossings of the 2nd derivative map
  // if a zero crossing does not occur, remove possible edge from the edge map
  for (i = 0; i < h; i++)
    for (j = 0; j < w; j++)
    {
      if (( i > 1) && (j > 1) && (i < h - 2) && (j < w - 2))
      {
        if (ABS(secondder[i][j] + secondder[i+1][j] + secondder[i][j+1]) ==
            ABS(secondder[i][j]) + ABS(secondder[i+1][j]) + ABS(secondder[i][j+1]))
          fEdgeMap[i][j] = 0;
      }
      else
      {
        fEdgeMap[i][j] = 0;
      }
    }

  if (P->saveall)
  {
    sprintf (fn, "output/edgemapimproved%c.pgm", index);
    writefloatpgm (fEdgeMap, w, h, fn, 0);
  }
  free_float_array (f, 5);
  free_float_array (secondder, h);

  return;
}


// =============================================================================
// TScene class methods
// =============================================================================

// =============================================================================
// TScene constructor
//
// Loads both images, and precalculates the matrices E, F, Ml, and Mr
// from the calibration data. 
//
// Parameters: char *n - name of the scene (for recognition of the image
//                       and calibration files)
// =============================================================================

TScene::TScene (char *n)
{
  char s[100];
  int i, j;

  Name = strdup (n);

  // read left and right image
  sprintf (s, "%sL.pgm", n);
  LImage = new TPGMImage (s, 'L');
  sprintf (s, "%sR.pgm", n);
  RImage = new TPGMImage (s, 'R');

  // read calibration data
  sprintf (s, "%sCAL.txt", n);
  CalFilename = strdup (s);
  getcalpars (CalFilename, &CParams);

  Cfirst = NULL; Ccurrent = NULL;
  ConF = NULL; ConC = NULL;

  T = vector (3);
  R = matrix (3,3);

  // ---------------------------------------------------------------------------
  // precalculate the matrices

  for (i = 1; i < 4; i++) T[i] = CParams.T[i-1];
  for (i = 1; i < 4; i++)
    for (j = 1; j < 4; j++)
    {
     R[i][j] = CParams.R[i-1][j-1];
    }

  S = matrix (3,3);
  S[2][1] = T[3]; S[1][2] = -T[3];
  printf("%f %f \n", T[3], S[2][1]);
  S[3][1] = -T[2]; S[1][3] = T[2];
  S[3][2] = T[1]; S[2][3] = -T[1];

  MLinv = matrix (3,3);
  MLinv[1][1] = CParams.leftcam.sx;
  MLinv[2][2] = CParams.leftcam.sy;
  MLinv[3][3] = 1;
  MLinv[1][3] = - CParams.leftcam.sx * CParams.leftcam.ox;
  MLinv[2][3] = - CParams.leftcam.sy * CParams.leftcam.oy;

  MRinv = matrix (3,3);
  MRinv[1][1] = CParams.rightcam.sx;
  MRinv[2][2] = CParams.rightcam.sy;
  MRinv[3][3] = 1;
  MRinv[1][3] = - CParams.rightcam.sx * CParams.rightcam.ox;
  MRinv[2][3] = - CParams.rightcam.sy * CParams.rightcam.oy;

  E = R * S;
  F = MRinv.transpose() * E * MLinv;

  return;
}

// =============================================================================
// TScene destructor
//
// Parameters: none
// Returns:    nothing
// =============================================================================

TScene::~TScene ()
{
  for (int a = 0; a < LImage->h; a++)
  {
    for (int b = 0; b < LImage->w; b++)
    {
      if (CorrespondenceMap[a][b] != NULL) delete CorrespondenceMap[a][b];
    }
    free (CorrespondenceMap[a]);
  }
  free (CorrespondenceMap);

  delete LImage;
  delete RImage;
  free (CalFilename);

  ConC = ConF;
  while (ConC != NULL)
  {
    ConC = ConC->next;
    delete ConF;
    ConF = ConC;
  }

  free (Name);
}

// =============================================================================
// void TScene::ExaminePoint (int i, int j, PCorrespondence *Cor,
//                            vector pOrigi, vector pTargi, matrix F,
//                            PPGMImage ISource, PPGMImage ITarget,
//                            PParams Params)
//
// Searches alongside the epipolar line for the best match. If best match is
// found and has a sufficient value of the correlation coefficient, the match
// is stored.
//
// Parameters: i, j    - coordinates of the point which is being matched
//             PCorrespondence *Cor  - position where to store the match
//             vector pOrigi, pTargi - source and target coordinates
//             matrix F - fundamental matrix
//             PPGMImage ISource, ITarget - images from and to which is
//                                          being matched
//             PParams Params - parameter structure
//
// Returns:    nothing
//
// =============================================================================

void TScene::ExaminePoint (int i, int j, PCorrespondence *Cor,
                           vector pOrigi, vector pTargi, matrix F,
                           PPGMImage ISource, PPGMImage ITarget,
                           PParams Params)
{
  vector pOrig(3), pTarg(3);     // coordinates in the image plane

  vector uTargi(3);  // vector for the equation of the epipolar line

  int a, b;
  int k, maxk;

  float x, y, dx, dy;
  int m, n, m1, m2, n1, n2;

  float initx, inity;

  float bestvalue, curvalue;
  int bestx, besty;

  int pixelcount;
  float avl, avr;
  float varl, varr;
  float corr, corrd, corra, corri;
  float distl, distr;

  uTargi = F * pOrigi;

  // ---------------------------------------------------------------------------
  // calculation of the slope of the epipolar line

  if (ABS(uTargi[1]) > ABS(uTargi[2]))
  {
    dy = 1;
    dx = - uTargi[2] / uTargi[1];
    x = -pTargi[3] * uTargi[3] / uTargi[1];
    y = 0;
    maxk = ITarget->h;
  }
  else if (uTargi[2] != 0)
  {
    dx = 1;
    dy = - uTargi[1] / uTargi[2];
    x = 0;
    y = -pTargi[3] * uTargi[3] / uTargi[2];
    maxk = ITarget->w;
  }
  else
  {
    dx = 0;
    dy = 1;
    x = -pTargi[3] * uTargi[3] / uTargi[1];
    y = 0;
    maxk = ITarget->h;
  }

  initx = x;
  inity = y;

  bestvalue = -1.0;
  bestx = -1;
  besty = -1;

  // ---------------------------------------------------------------------------
  // now search along the epipolar line
  for (k = 0; k < maxk; k++)
  {
    // calculate the position of the pixel which is to be matched
    pTarg[1] = floor(x + 0.5);
    pTarg[2] = floor(y + 0.5);
    a = (int) pTarg[2];
    b = (int) pTarg[1];
    if ((a >= 0) && (a < ITarget->h) &&
        (b >= 0) && (b < ITarget->w))
    {

    // uncomment if the epipolar line of a desired pixel should be drawn
    // into the image
//     if ((i == 399) && (j == 399))

// ITarget->Data[a][b] = 255;

      // if we are on an edge pixel, we can start

      if (ITarget->EdgeMap[a][b] > 0)
      {

        // first, determine the area in which the matching pixel should be
        // looked up (usually 0 or +/-1 pixel)
        if (a > 0) a = a - Params->lookaround;

        while (a <= MIN((int) pTarg[2] + Params->lookaround, ITarget->h - 1))
        {
          if (b > 0) b = b - Params->lookaround;

          while (b <= MIN((int) pTarg[1] + Params->lookaround, ITarget->w - 1))
          {

            // now calculate the similarity measure

            // precalculate the window position so that it does not run
            // of the image
            m1 = MAX (-(Params->corrarea), -i); m1 = MAX (m1, -a);
            m2 = MIN (Params->corrarea, ISource->h - i - 1);
            m2 = MIN (m2, ITarget->h - a - 1);
            n1 = MAX (-(Params->corrarea), -j); n1 = MAX (n1, -b);
            n2 = MIN (Params->corrarea, ISource->w - j - 1);
            n2 = MIN (n2, ITarget->w - b - 1);

            // -----------------------------------------------------------------
            // first correlation based on the edge intensities
            pixelcount = 0; avl = 0.0; avr = 0.0;
            varl = 0.0; varr = 0.0;
            corr = 0.0;

            for (m = m1; m <= m2; m++)
            {
              for (n = n1; n <= n2; n++)
              {
                pixelcount++;
                avl += ISource->fEdgeMap[i + m][j + n];
                avr += ITarget->fEdgeMap[a + m][b + n];
              }
            }
            avl = avl / (float) pixelcount;
            avr = avr / (float) pixelcount;

            for (m = m1; m <= m2; m++)
            {
              for (n = n1; n <= n2; n++)
              {
                corr += (ISource->fEdgeMap[i + m][j + n] - avl) *
                        (ITarget->fEdgeMap[a + m][b + n] - avr);
                varl += (ISource->fEdgeMap[i + m][j + n] - avl) *
                        (ISource->fEdgeMap[i + m][j + n] - avl);
                varr += (ITarget->fEdgeMap[a + m][b + n] - avr) *
                        (ITarget->fEdgeMap[a + m][b + n] - avr);
              }
            }

            if ((varl > 0) && (varr > 0))
            {
              corrd = corr / powl (varl, 0.5) / powl (varr, 0.5);
            }
            else
            {
              corrd = -1000;
            }

            // -----------------------------------------------------------------
            // now correlation in the direction map
            pixelcount = 0; avl = 0.0; avr = 0.0;
            varl = 0.0; varr = 0.0;
            corr = 0.0;

            for (m = m1; m <= m2; m++)
            {
              for (n = n1; n <= n2; n++)
              {
                pixelcount++;
                avl += ISource->fDirectionMap[i + m][j + n];
                avr += ITarget->fDirectionMap[a + m][b + n];
              }
            }
            avl = avl / (float) pixelcount;
            avr = avr / (float) pixelcount;

            for (m = m1; m <= m2; m++)
            {
              for (n = n1; n <= n2; n++)
              {
                distl = ABS(ISource->fDirectionMap[i + m][j + n] - avl);
                distl = MIN(distl, M_PI - distl);
                distr = ABS(ITarget->fDirectionMap[a + m][b + n] - avr);
                distr = MIN(distr, M_PI - distr);
                corr += distl * distr;
                varl += distl * distl;
                varr += distr * distr;
              }
            }

            if ((varl > 0) && (varr > 0))
            {
              corra = corr / powl (varl, 0.5) / powl (varr, 0.5);
            }
            else
            {
              corra = -1000;
            }

            // -----------------------------------------------------------------
            // correlation in the actual picture (intensities)
            pixelcount = 0; avl = 0.0; avr = 0.0;
            varl = 0.0; varr = 0.0;
            corr = 0.0;

            for (m = m1; m <= m2; m++)
            {
              for (n = n1; n <= n2; n++)
              {
                pixelcount++;
                avl += ISource->fData[i + m][j + n];
                avr += ITarget->fData[a + m][b + n];
              }
            }
            avl = avl / (float) pixelcount;
            avr = avr / (float) pixelcount;

            for (m = m1; m <= m2; m++)
            {
              for (n = n1; n <= n2; n++)
              {
                corr += (ISource->fData[i + m][j + n] - avl) *
                        (ITarget->fData[a + m][b + n] - avr);
                varl += (ISource->fData[i + m][j + n] - avl) *
                        (ISource->fData[i + m][j + n] - avl);
                varr += (ITarget->fData[a + m][b + n] - avr) *
                        (ITarget->fData[a + m][b + n] - avr);
              }
            }

            if ((varl > 0) && (varr > 0))
            {
              corri = corr / powl (varl, 0.5) / powl (varr, 0.5);
            }
            else
            {
              corri = - 1000;
            }

            // -----------------------------------------------------------------
            // now calculate some (empirically estimated) combinations
            // that work nicely
            corr = (corrd + corra + 0 * corri) / 2;
            if (corr > bestvalue)
            {
              bestvalue = corr;
              bestx = b;
              besty = a;
            }

            corr = (corrd + corra + 1 * corri) / 3;
            if (corr > bestvalue)
            {
              bestvalue = corr;
              bestx = b;
              besty = a;
            }
            corr = (2 * corrd + 1 * corra + 1 * corri) / 4;
            if (corr > bestvalue)
            {
              bestvalue = corr;
              bestx = b;
              besty = a;
            }

            b++;
          }  // (while b <= (int) pr[2] + 1)

          b = b - 1 - Params->lookaround;
          a++;
        } // (while a <= (int) pr[2] + 1)

      }
    }
    x = x + dx;
    y = y + dy;
  }

  // ---------------------------------------------------------------------------
  // now, if the best match satisfies the correlation coefficient threshold,
  // save the match

  if (bestvalue > Params->corrthreshold)
  {
    if (Params->showcorrs)
      printf ("point: %d %d match %d %d bestvalue %f\n",
              j, i, bestx, besty, bestvalue);
    if (*Cor == NULL)
    {
      *Cor = new TCorrespondence;
    }

    (*Cor)->xli = j;
    (*Cor)->yli = i;
    (*Cor)->xri = bestx;
    (*Cor)->yri = besty;

    // here, signs had to be changed as well (compared to the slides)

    (*Cor)->xl = (j - CParams.leftcam.ox) * CParams.leftcam.sx;
    (*Cor)->yl = (i - CParams.leftcam.oy) * CParams.leftcam.sy;
    (*Cor)->xr = (bestx - CParams.rightcam.ox) * CParams.rightcam.sx;
    (*Cor)->yr = (besty - CParams.rightcam.oy) * CParams.rightcam.sy;

    (*Cor)->initx = initx;
    (*Cor)->inity = inity;
    (*Cor)->corr = bestvalue;

    (*Cor)->next = NULL;

  }

  return;
}

// =============================================================================
// void TScene::FindCorrespondences(PParams Params)
//
// This method looks for correspondences for all edge points in the edge map.
// For each edge point, the method ExaminePoint is called.
// Two approaches are implemented - taking all edge points, or submitting the
// desired edge points in a file.
//
// Parameters: PParams Params - parameter structure
//
// Returns:    nothing
//
// =============================================================================

void TScene::FindCorrespondences(PParams Params)
{
  int i, j;
  int pointx, pointy;
  int lastx, lasty;

  vector pli(3), pri(3);

  PCorrespondence Cor;

  // allocate memory for the correspondences

  CorrespondenceMap = (PCorrespondence **) calloc(LImage->h, sizeof(PCorrespondence *));
  for (i = 0; i < LImage->h; i++)
  {
    CorrespondenceMap[i] = (PCorrespondence *) calloc(LImage->w, sizeof(PCorrespondence));
    for (j = 0; j < LImage->w; j++)
      CorrespondenceMap[i][j] = NULL;
  }

  // ---------------------------------------------------------------------------

  if (Params->pointsgiven == 0)
  {
    // in this case, take all edge points from the edge map

    for (i = 0; i < LImage->h; i++)
    {
      printf (".");
      for (j = 0; j < LImage->w; j++)
      {
        if (LImage->EdgeMap[i][j] > 0)  // edge found
        {

          pli[3] = CParams.leftcam.f;
          pri[3] = CParams.rightcam.f;

          pli[1] = j;
          pli[2] = i;

          // call the matching algorithm
          ExaminePoint (i, j, &(CorrespondenceMap[i][j]), pli, pri, F,
                        LImage, RImage, Params);

          // if matching was successful, call reverse matching

          if (CorrespondenceMap[i][j] != NULL)
          {
            Cor = NULL;
            pri[1] = CorrespondenceMap[i][j]->xri;
            pri[2] = CorrespondenceMap[i][j]->yri;
            ExaminePoint (CorrespondenceMap[i][j]->yri,
                          CorrespondenceMap[i][j]->xri, &Cor,
                          pri, pli, F.transpose(), RImage, LImage,
                          Params);
            if (Cor != NULL)
            {
              if ((ABS(Cor->yri - (float)i) <= Params->errorallowed) &&
                  (ABS(Cor->xri - (float)j) <= Params->errorallowed))
              {
                // if the difference for the reversed match is sufficiently low,
                // confirm the match
                CorrespondenceMap[i][j]->valid = 1;

              }
              else
              {
                // otherwise remove it
                delete CorrespondenceMap[i][j];
                CorrespondenceMap[i][j] = NULL;
              }

              delete Cor;
            }
            else
            {
              delete CorrespondenceMap[i][j];
              CorrespondenceMap[i][j] = NULL;
            }
          }


        } // if
      }  // for j
    }  // for i
  }
  else
  {
    // -------------------------------------------------------------------------
    // in this case, points are submitted from a file
    FILE *fp;

    fp = fopen (Params->pointsfile, "r");
    if (!fp) {
	fprintf (stderr, "Error opening file containing the points.\n");
	exit (0);
    }

    printf ("Points:\n");
    lastx = -1; lasty = -1;

    while (!feof (fp))
    {
      // read all points from the file

      if (fscanf (fp, "%d %d\n", &pointx, &pointy) == 2)
      {
        if ((pointx >= 0) && (pointx < LImage->w) &&
            (pointy >= 0) && (pointy < LImage->h))
        {
          pli[3] = CParams.leftcam.f;
          pri[3] = CParams.rightcam.f;

          pli[1] = pointx;
          pli[2] = pointy;

          // -------------------------------------------------------------------
          // call the matching algorithm

          ExaminePoint (pointy, pointx, &(CorrespondenceMap[pointy][pointx]),
                        pli, pri, F, LImage, RImage, Params);

          // if match is found, run the reversed matching

          if (CorrespondenceMap[pointy][pointx] != NULL)
          {
            Cor = NULL;
            pri[1] = CorrespondenceMap[pointy][pointx]->xri;
            pri[2] = CorrespondenceMap[pointy][pointx]->yri;
            ExaminePoint (CorrespondenceMap[pointy][pointx]->yri,
                          CorrespondenceMap[pointy][pointx]->xri, &Cor,
                          pri, pli, F.transpose(), RImage, LImage,
                          Params);
            if (Cor != NULL)
            {
              printf ("ORig: %d %d, Target %d %d, backwards: %d %d -> %d %d\n",
                      pointx, pointy,
                      (int)(CorrespondenceMap[pointy][pointx]->xri),
                      (int)(CorrespondenceMap[pointy][pointx]->yri),
                      (int)pri[1], (int)pri[2], (int)(Cor->xri), (int)(Cor->yri));
              if ((ABS(Cor->yri - (float)pointy) <= Params->errorallowed) &&
                  (ABS(Cor->xri - (float)pointx) <= Params->errorallowed))
              {
                // match is confirmed by the reversed matching algorithm
                CorrespondenceMap[pointy][pointx]->valid = 1;
              }
              else
              {
                // match is not confirmed - delete it
                delete CorrespondenceMap[pointy][pointx];
                CorrespondenceMap[pointy][pointx] = NULL;
              }

              delete Cor;
            }
            else
            {
              delete CorrespondenceMap[pointy][pointx];
              CorrespondenceMap[pointy][pointx] = NULL;
            }
          }


          // -------------------------------------------------------------------
          // keep track of connections between the corner points

          if (CorrespondenceMap[pointy][pointx] != NULL)
          {

            if (lastx != -1)
            {
              if (ConF == NULL)
              {
                ConF = new TConnection;
                ConC = ConF;
              }
              else
              {
                ConC->next = new TConnection;
                ConC = ConC->next;
              }
              ConC->next = NULL;
              ConC->i1 = lasty;
              ConC->j1 = lastx;
              ConC->i2 = pointy;
              ConC->j2 = pointx;

            }
            lastx = pointx;
            lasty = pointy;

          }
          else
          {
            // if no match is found, the point is refused
            printf ("  examined %d %d  ", pointx, pointy);
            printf ("NOT ACCEPTED\n");
            lastx = -1;
            lasty = -1;
          }

        }
      }
      else
      {
        // # in the file means start of a new shape (no line between the
        // previous and next point

        fscanf(fp, "#\n");
        printf ("New shape:\n");
        lastx = -1; lasty = -1;
      }
    }


  }


  return;
}

// =============================================================================
// void TScene::Triangulate (PParams Params)
//
// Calculates the 3-D reconstruction of the scene using triangulation.
//
// Parameters: PParams Params - parameter structure
//
// Returns:    nothing
//
// =============================================================================

void TScene::Triangulate (PParams Params)
{
  vector pars(3), pos(3), pos2(3);
  matrix m(3, 3);
  vector pl(3), pr(3), v(3), w(3);

  INT32 i = 0;

  int a, b, c, d;

  int count;
  float av;

  // ---------------------------------------------------------------------------
  // run through all found correspondences

  for (a = 0; a < LImage->h; a++)
  {
    for (b = 0; b < LImage->w; b++)
    {
      Ccurrent = CorrespondenceMap[a][b];
      if (Ccurrent != NULL)
      {
        // construct matrix for solving the triangulation problem
        // first column is pl (corresponds to parameter a)
        pl[1] = Ccurrent->xl;
        pl[2] = Ccurrent->yl;
        pl[3] = CParams.leftcam.f;
        m[1][1] = pl[1];
        m[2][1] = pl[2];
        m[3][1] = pl[3];
        // second column is -R'pr (corresponds to parameter b)
        pr[1] = Ccurrent->xr;
        pr[2] = Ccurrent->yr;
        pr[3] = CParams.rightcam.f;

        v = R.transpose() * pr;

        m[1][2] = -v[1];
        m[2][2] = -v[2];
        m[3][2] = -v[3];

        // third column is pl x R'pr
        w = pl ^ v;
        m[1][3] = w[1];
        m[2][3] = w[2];
        m[3][3] = w[3];

        // ---------------------------------------------------------------------
        // calculate the vector pars, which contains the parameters a, b, c

        pars = m.inverse() * T;

        if (pars != vector (3))
        {
          // set of parameters found, calculate the 3-D position of the point
          // (from the point of view of the left camera)

          pos = pars[1] * pl + 0.5 * pars[3] * w;

          pos2 = pars[2] * v + T - 0.5 * pars[3] * w;

          Ccurrent->X = pos[1];
          Ccurrent->Y = pos[2];
          Ccurrent->Z = pos[3];
          Ccurrent->diff = (pars[3] * w).abs();

          Ccurrent->valid = 1;
        }
        else
        {
          // the inverse could not be calculated - drop the point
          Ccurrent->valid = 0;
          printf("Invalid parameter set. Ignoring.\n");
        }

        // if the point does not lie in the selected range (in case that any
        // range is chosen), get rid of it

        if ((Params->restrictcoordinates) &&
            ((Ccurrent->X < Params->x1) || (Ccurrent->X > Params->x2) ||
             (Ccurrent->Y < Params->y1) || (Ccurrent->Y > Params->y2) ||
             (Ccurrent->Z < Params->z1) || (Ccurrent->Z > Params->z2)))
        {
          delete CorrespondenceMap[a][b];
          CorrespondenceMap[a][b] = NULL;

        }

        i++;
      }
    } // for b

  }

  // now refine the coordinates (reduce the variance by averaging
  // the Z coordinates of pixels lying next to each other)

  for (a = 0; a < LImage->h; a++)
  {
    for (b = 0; b < LImage->w; b++)
    {
      Ccurrent = CorrespondenceMap[a][b];
      if (Ccurrent != NULL)
      {
        av = 0;
        count = 0;
        for (c = MAX (0, a - 1); c <= MIN (LImage->h - 1, a + 1); c++)
        {
          for (d = MAX (0, b - 1); d <= MIN (LImage->w - 1, b + 1); d++)
          {
            if ((CorrespondenceMap[c][d] != NULL) &&
                (ABS (CorrespondenceMap[c][d]->xri - CorrespondenceMap[a][b]->xri) < 3) &&
                (ABS (CorrespondenceMap[c][d]->yri - CorrespondenceMap[a][b]->yri) < 3))
            {
              av = av + CorrespondenceMap[c][d]->Z;
              count++;
            }
          }
        }
        CorrespondenceMap[a][b]->Z = av / count;
      }
    }

  }

  return;
}

// =============================================================================
// void TScene::SaveCoordinates (PParams Params)
//
// Saves the coordinates into the required files.
//
// Parameters: PParams Params - parameter structure
//
// Returns:    nothing
//
// =============================================================================

void TScene::SaveCoordinates (PParams Params)
{
  FILE *fcorr, *f3D;
  char fn[40];
  int a, b;

  PCorrespondence c1, c2;

  sprintf (fn, "output/%scorr.txt", Name);
  if ((fcorr = fopen (fn, "w")) == NULL)
  {
    printf ("Could not open %s.\n", fn);
    return;
  }
  sprintf (fn, "output/%s3d.txt", Name);
  if ((f3D = fopen (fn, "w")) == NULL)
  {
    printf ("Could not open %s.\n", fn);
    return;
  }
  printf ("\nSaving only restricted coordinates: x %f %f  y %f %f  z %f %f\n",
           Params->x1, Params->x2, Params->y1, Params->y2,
           Params->z1, Params->z2);

  for (a = 0; a < LImage->h; a++)
  {
    for (b = 0; b < LImage->w; b++)
    {
      Ccurrent = CorrespondenceMap[a][b];
      if (Ccurrent != NULL)
      {

        if (Ccurrent->valid)
        {
          if ((!Params->restrictcoordinates) ||
              ((Ccurrent->X >= Params->x1) && (Ccurrent->X <= Params->x2) &&
               (Ccurrent->Y >= Params->y1) && (Ccurrent->Y <= Params->y2) &&
               (Ccurrent->Z >= Params->z1) && (Ccurrent->Z <= Params->z2)))
          {
            fprintf (fcorr, "%d %d %d %d\n",
                     (int)Ccurrent->xli, (int)Ccurrent->yli,
                     (int)Ccurrent->xri, (int)Ccurrent->yri);

            fprintf (f3D, "%f %f %f\n",
                     Ccurrent->X * Params->swapx, Ccurrent->Y * Params->swapy,
                     Ccurrent->Z);

            LImage->Data[(UINT32)Ccurrent->yli]
                        [(UINT32)Ccurrent->xli] = 255;
            RImage->Data[(UINT32)Ccurrent->yri]
                        [(UINT32)Ccurrent->xri] = 255;

          }
        }
      } // if
    } // for b
  }

  // ---------------------------------------------------------------------------
  // now draw lines between points, if this was required from the command
  // line

  float curx, cury, curz;
  int count = 1;

  ConC = ConF;

  while (ConC != NULL)
  {
    if ((Params->connectpoints) &&
        (CorrespondenceMap[ConC->i1][ConC->j1] != NULL) &&
        (CorrespondenceMap[ConC->i2][ConC->j2] != NULL))
    {

      c1 = CorrespondenceMap[ConC->i1][ConC->j1];
      c2 = CorrespondenceMap[ConC->i2][ConC->j2];
      count = 0;
      if ((c1->valid) && (c2->valid))
      {
        // interpolate between the two points
        while (count <= 100)
        {
          curx = c1->X + (c2->X - c1->X)*((float) count) / 100.0;
          cury = c1->Y + (c2->Y - c1->Y)*((float) count) / 100.0;
          curz = c1->Z + (c2->Z - c1->Z)*((float) count) / 100.0;

          fprintf (f3D, "%f %f %f\n",
                   curx * Params->swapx, cury * Params->swapy, curz);

          count++;
        }

      }

    }
    ConC = ConC->next;
  }

  fclose (fcorr);
  fclose (f3D);

  writepgm (LImage->Data, LImage->w, LImage->h, "output/testL.pgm");
  writepgm (RImage->Data, RImage->w, RImage->h, "output/testR.pgm");

  return;
}


// =============================================================================
// PParams ProcessParameters(int argc, char* argv[])
//
// Processes command-line parameters, prints help.
//
// Parameters: int argc     - number of command-line parameters
//             char* argv[] - array containing command-line parameters
// Returns:    processed command line parameters
// =============================================================================

PParams ProcessParameters(int argc, char* argv[])
{
  PParams P;
  int argindex;

  // ---------------------------------------------------------------------------
  // print help
  if (argc > 1)
  {
    if ((strcmp(argv[1], "-h")== 0) || (strcmp(argv[1], "-?")== 0))
    {
       printf ("\n");
       printf ("Usage: stereo <scenename> [parameters]\n");
       printf ("       stereo -h or stereo -?    get this help\n");
       printf ("Parameters:\n");
       printf ("  -t value     set edge threshold to 'value' (default 50)\n");
       printf ("  -s value     smooth the image first, size of mask is (2*value+1)\n");
       printf ("  -d           use 2nd derivative filter to improve edge look-up\n");
       printf ("  -saveall     save all intermediate files (incl.edge and direction map)\n");
       printf ("  -swapx       flip the x coordinate upside-down\n");
       printf ("  -swapy       flip the y coordinate upside-down\n");
       printf ("  -r x1 x2 y1 y2 z1 z2  restrict the scene to the given block\n");
       printf ("  -points filename      submit point positions from file\n");
       printf ("  -connect              connect the submitted points with lines\n");
       printf ("  -lookaround n     search for matches in distance n from epipolar line\n");
       printf ("  -corrarea W       set the size of correlation window to 2W+1\n");
       printf ("  -corrthreshold c  minimum value for the correlation coefficient\n");
       printf ("  -error n          allowed error in pixels for reversed matching\n");
       printf ("  -showcorrs        show calculated correlation coefficients\n");
       printf ("  -h | -?      get this help\n\n");

       exit(0);
    }
  }

  if (argc < 2)
  {
    printf ("Not enough parameters. Run 'stereo -h' to get help.\n\n");
    exit(0);
  }

  // ---------------------------------------------------------------------------
  // process parameters

  P = (PParams) calloc(1, sizeof(TParams));

  // default values
  P->contthreshold = 50;
  P->smoothimage = 0;
  P->saveall = 0;
  P->swapx = 1;
  P->swapy = 1;
  P->pointsgiven = 0;
  P->lookaround = 0;

  P->restrictcoordinates = 0;
  P->corrarea = 10;
  P->corrthreshold = 0.5;
  P->errorallowed = 0;
  P->showcorrs = 0;

  // get the name of the scene from the first command-line argument
  strcpy (P->scenename = (char *) malloc(strlen(argv[1])+1), argv[1]);

  argindex = 2;

  // loop for parameter processing
  while (argindex < argc)
  {
    // threshold in the contour map
    if ((strcmp (argv[argindex],"-t") == 0) && (argindex < argc - 1))
    {
      if (atol(argv[argindex+1]) > 0)
      {
        P->contthreshold = atol(argv[argindex+1]);
      }
      argindex++;
    }
    else if ((strcmp (argv[argindex],"-s") == 0) && (argindex < argc - 1))
    {
      if (atol(argv[argindex+1]) > 0)
      {
        P->smoothimage = atol(argv[argindex+1]);
      }
      argindex++;
    }
    else if (strcmp (argv[argindex],"-saveall") == 0)
    {
      P->saveall = 1;
    }
    else if (strcmp (argv[argindex],"-d") == 0)
    {
      P->use2ndderivative = 1;
    }
    else if (strcmp (argv[argindex],"-swapx") == 0)
    {
      P->swapx = -1;
    }
    else if (strcmp (argv[argindex],"-swapy") == 0)
    {
      P->swapy = -1;
    }
    // restrict the region (xmin, xmax, ymin, ymax, zmin, zmax)
    else if ((strcmp (argv[argindex],"-r") == 0) && (argindex < argc - 6))
    {
      P->restrictcoordinates = 1;
      P->x1 = atol(argv[argindex+1]);
      P->x2 = atol(argv[argindex+2]);
      P->y1 = atol(argv[argindex+3]);
      P->y2 = atol(argv[argindex+4]);
      P->z1 = atol(argv[argindex+5]);
      P->z2 = atol(argv[argindex+6]);

      argindex = argindex + 6;
    }
    else if ((strcmp (argv[argindex],"-points") == 0) && (argindex < argc - 1))
    {
      P->pointsgiven = 1;
      P->pointsfile = strdup (argv[argindex+1]);
      argindex++;
    }
    else if ((strcmp (argv[argindex],"-lookaround") == 0) && (argindex < argc - 1))
    {
      if (atol(argv[argindex+1]) > 0)
      {
        P->lookaround = atol(argv[argindex+1]);
      }
      argindex++;
    }
    else if (strcmp (argv[argindex],"-connect") == 0)
    {
      P->connectpoints = 1;
    }
    else if ((strcmp (argv[argindex],"-corrarea") == 0) && (argindex < argc - 1))
    {
      if (atol(argv[argindex+1]) > 0)
      {
        P->corrarea = atol(argv[argindex+1]);
      }
      argindex++;
    }
    else if ((strcmp (argv[argindex],"-corrthreshold") == 0) && (argindex < argc - 1))
    {
      P->corrthreshold = atof(argv[argindex+1]);
      argindex++;
    }
    else if ((strcmp (argv[argindex],"-error") == 0) && (argindex < argc - 1))
    {
      P->errorallowed = atol(argv[argindex+1]);
      argindex++;
    }
    else if (strcmp (argv[argindex],"-showcorrs") == 0)
    {
      P->showcorrs = 1;
    }

    argindex = argindex + 1;
  }

  return P;
}

// =============================================================================
// void writefloatpgm (float **im, int w, int h, char *fname, float shift)
//
// converts the array of floats into an array of unsigned chars,
// normalising the values into the interval 0..255, and
// writes the array of unsigned chars into a .pgm file
//
// Parameters: float **im   - array to be converted
//             int w, int h - width and height of the array (image)
//             char *fname  - filename to be used
//             float shift  - value to be added to every pixel
// Returns:    nothing
// =============================================================================

void writefloatpgm (float **im, int w, int h, char *fname, float shift)
{
  unsigned char **imch;
  int i, j;
  float maxval;

  maxval = 0;
  imch = uchar_array (w, h);

  // find highest and smallest value (if negative values occur)
  for (i = 0; i < h; i++)
    for (j = 0; j < w; j++)
    {
      if (ABS(im[i][j] + shift) > maxval) maxval = ABS(im[i][j] + shift);
    }

  // normalise
  if (maxval > 0)
    for (i = 0; i < h; i++)
      for (j = 0; j < w; j++)
        imch[i][j] = (char)floor(ABS(im[i][j] + shift)/maxval*255);

  writepgm (imch, w, h, fname);

  free_uchar_array (imch, h);

  return;
}

// =============================================================================
// float **convolution (float **image, int iw, int ih,
//                      float **filter, int fw, int fh)
//
// Calculates the discrete 2D-convolution of the two arrays: image and filter.
// The coordinates of the filter are centered (i.e. (0,0) in the centre of
// the filter). It is therefore adviceable (but not necessary)
// to use filters with odd height and width.
//
// Parameters: float **image  - array containing the image
//             int iw, int ih - width and height of the image
//             float **filter - array containing the filter
//             int fw, int fh - width and height of the filter
// Returns:    pointer to the convoluted array
// =============================================================================

float **convolution (float **image, int iw, int ih,
                     float **filter, int fw, int fh)
{
  float **output;
  int i, j, k, l;

  // Allocate memory for the output array
  output = float_array (iw, ih);

  for (k = 0; k < ih; k++)
  {
    for (l = 0; l < iw; l++)
    {
      output[k][l] = 0;
      for (i = MAX(k-(fh-1)/2,0); i <= MIN(k+fh/2,ih-1); i++)
      {
        for (j = MAX(l-(fw-1)/2,0); j <= MIN(l+fw/2,iw-1); j++)
        {
          output[k][l] += (image[i][j])*(filter[k-i+fh/2][l-j+fw/2]);
        }
      }

    }
  }

  return output;
}

// =============================================================================
// float **smoothimage (float **im, int w, int h, PParams P)
//
// Smoothes the image according to the value in P->smoothimage.
// The mask of the smoothing filter will have size 2*(P->smoothimage)+1.
//
// Parameters: float **im    - image to be smoothed
//             int w, int h  - width and height of the image
//             PParams P     - command-line parameters
// Returns:    pointer to the smoothed image
// =============================================================================

float **smoothimage (float **im, int w, int h, PParams P)
{
  float **output, **smoothfilter;
  int i, j, s;
  float sum;

  s = P->smoothimage;
  if (s)
  {
    printf ("Smoothing the image.\n");

    smoothfilter = float_array (2*s+1, 2*s+1);

    // define the smoothing filter
    sum = 0;
    for (i = 0; i < 2*s+1; i++)
      for (j = 0; j < 2*s+1; j++)
      {
        smoothfilter[i][j] = 1/(2*M_PI*s*s) *
          exp ( -0.5*((i-s)*(i-s)+(j-s)*(j-s))/(s*s) );
        sum += smoothfilter[i][j];
      }

    // normalise so that sum of the values is 1
    for (i = 0; i < 2*s+1; i++)
      for (j = 0; j < 2*s+1; j++)
        smoothfilter[i][j] /= sum;


    // convoluting the image with the filter
    output = convolution (im, w, h, smoothfilter, 2*s+1, 2*s+1);

    free_float_array (im, h);
    free_float_array(smoothfilter, 3);

    // save the smoothed file, if requested
//    if (P->saveall) writefloatpgm (output, w, h, "_smoothpicture.pgm", 0);
  }
  else
  {
    output = im;
  }

  return output;
}

// =============================================================================
// int main(int argc, char* argv[])
//
// Parameters : command-line parameters
// Returns:     0 on success, 1 on failure
// =============================================================================

int main(int argc, char* argv[])
{

  PParams Params;

  Params = ProcessParameters (argc, argv);

  PScene Scene = new TScene(Params->scenename);

  Scene->LImage->GetEdgeMap(Params);
  Scene->RImage->GetEdgeMap(Params);

  Scene->FindCorrespondences(Params);
  Scene->Triangulate(Params);

  Scene->SaveCoordinates(Params);

  // clean-up memory

  delete Scene;

  free (Params->scenename);
  free (Params);

  return 0;

}

